/**
  Rendering demo application.
  
  Orion Sky Lawlor, olawlor@acm.org, 2010-04-08 (Public Domain)
*/
#include <iostream>
#include <cmath>
#include <GL/glew.h> /* 
OpenGL Extensions Wrangler:  for gl...ARB extentions.  Must call glewInit after glutCreateWindow! */
#include <GL/glut.h> /* OpenGL Utilities Toolkit, for GUI tools */
#include "ogl/glsl.h"
#include "ogl/glsl.cpp"
#include "ogl/minicam.h"
#include "osl/mat4.h"
#include "osl/mat4_inverse.h"

#include <vector>


class triangle {
public:
	int a[3]; // 1-based index into points and normals array
	triangle(int *a_) {
		for (int i=0;i<3;i++) a[i]=a_[i];
	}
};

void display(void) 
{
	glEnable(GL_DEPTH_TEST);
	glDepthMask(GL_TRUE); // turn on Z writes
	glEnable(GL_BLEND);
	
	glClearColor(0.4,0.5,0.7,0.0); // background color
	glClear(GL_COLOR_BUFFER_BIT + GL_DEPTH_BUFFER_BIT);
	
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity(); // flush any ancient matrices
	gluPerspective(70.0, // fov
		glutGet(GLUT_WINDOW_WIDTH)/(float)glutGet(GLUT_WINDOW_HEIGHT),
		0.1,
		1000.0); // z clipping planes
	float cameraVelocity=3.0;
	oglCameraLookAt(cameraVelocity);
	
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity(); // flush any ancient matrices
	
	float time_uniform=glutGet(GLUT_ELAPSED_TIME)*0.001;
	
/* Upload 3D noise texture, to texture unit 2 */
	static GLuint noisetex=0;
	if (noisetex==0) { /* only need to build and upload one time; it stays set up. */
		glGenTextures(1,&noisetex);
		glBindTexture(GL_TEXTURE_3D,noisetex);
		int sz=32; /* RGBA 32 x 32 x 32 */
		float *noise=new float[4*sz*sz*sz];
		for (int z=0;z<sz;z++) 
		for (int y=0;y<sz;y++) 
		for (int x=0;x<sz;x++) 
		for (int chan=0;chan<4;chan++) {
			float r=(rand()&0xfff)*(1.0/0xfff);
			noise[chan+4*(x+sz*(y+sz*z))]=r;
		}
		glTexImage3D(GL_TEXTURE_3D,0,GL_RGBA8,
			sz,sz,sz,0,
			GL_RGBA,GL_FLOAT,noise);
		glTexParameteri(GL_TEXTURE_3D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
		glTexParameteri(GL_TEXTURE_3D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
		delete[] noise;
		glBindTexture(GL_TEXTURE_3D,0);
	}
	glActiveTexture(GL_TEXTURE2);
	glBindTexture(GL_TEXTURE_3D,noisetex);
	glActiveTexture(GL_TEXTURE0);
	glBindTexture(GL_TEXTURE_3D,0);
	
/* Load and draw the cat model */
	GLhandleARB catshader=MAKE_PROGRAM_OBJECT(
		varying vec3 position; // location of fur-deformed geometry
		varying vec3 texcoords; // location of *original* geometry
		varying vec4 fg;
		varying vec4 bg; // vec4 colors (w/alpha)  
		varying vec3 normal; // surface normal
		void main(void) {
			vec4 world=gl_ModelViewMatrix*gl_Vertex;
			position=vec3(world);
			normal=vec3(gl_NormalMatrix*gl_Normal);
			gl_Position=gl_ProjectionMatrix*world;
			texcoords=vec3(gl_ModelViewMatrix*gl_MultiTexCoord0);
			float fur=gl_MultiTexCoord0.w;
			float a=1.0-fur;
			fg=vec4(0.9,0.5,0.3,a); // light wood
			bg=vec4(0.3,0.2,0.1,a*0.2); // dark wood
			if (fur==0) {/* first layer */
				fg.a=bg.a=1.0;
			}
		}
	,
		varying vec3 position; // location of fur-deformed geometry
		varying vec3 texcoords; // location of *original* geometry
		varying vec4 fg;
		varying vec4 bg; // vec4 colors (w/alpha)
		varying vec3 normal; // surface normal
		uniform float time_uniform;
		
		uniform sampler3D noisetex;
		// Return value noise at this location
		vec4 mynoise(vec3 loc) {
			return texture3D(noisetex,loc);
		}
		
		// Return the color of the solid texture at this point
		vec4 multinoise(vec3 pos) {
			vec4 octave0=mynoise(pos);
			vec4 octave1b=mynoise(pos*3.0);
			vec4 octave1=mynoise(pos*10.0);
			vec4 octave2b=mynoise(pos*30.0);
			vec4 octave2=mynoise(pos*100.0);
			return (0.3*octave0+0.3*octave1b+0.2*octave1+0.1*octave2b+0.1*octave2);
		}
		
		// Cellular texture
		vec4 cellular(vec3 pos) {
			vec3 cell=floor(pos);
			float closest=10.0;
			for (float dz=-1.0;dz<=1.0;dz++)
			for (float dy=-1.0;dy<=1.0;dy++)
			for (float dx=-1.0;dx<=1.0;dx++) {
				vec3 curcell=cell+vec3(dx,dy,dz);
				vec3 noiseloc=curcell*0.353674+0.01*vec3(time_uniform);
				vec3 randpt=curcell+1.4*vec3(mynoise(noiseloc));
				float cur=length(pos-randpt);
				closest=min(closest,cur);
			}
			vec4 fg=vec4(0.9,0.9,0.9,1.0); // light near center
			vec4 bg=vec4(0.3,0.2,0.1,1.0); // dark far away
			return bg+closest*(fg-bg);
		}
		
		// Return fur color
		vec4 furtexture(vec3 pos) {
			float blend=(mynoise(pos*0.3).x-0.5)*2.0;
			vec4 ret=bg+clamp(blend,0.0,1.0)*(fg-bg); // clamp color blend
			ret.a=bg.a+blend*(fg.a-bg.a); // don't clamp alpha blend--leaves spaces between hairs
			return ret;
		}
		
		void main(void) {
			vec4 color=furtexture(texcoords*10.0);
		
			// Lighting calculation
			const vec3 L=normalize(vec3(0.4,0.0,1.0));
			vec3 N=normalize(normal);
			float d=clamp(dot(N,L),0.0,0.8)+0.2;
			gl_FragColor=color*vec4(vec3(d),1.0);
		}
	);
	glUseProgramObjectARB(catshader);
	glFastUniform1i(catshader,"noisetex",2); // texture unit 2
	glFastUniform1f(catshader,"time_uniform",time_uniform);
	
	
	std::vector<vec3> points; points.push_back(vec3(-999.0));
	std::vector<vec3> normals; normals.push_back(vec3(-999.0));
	std::vector<vec3> texcoords; texcoords.push_back(vec3(-999.0));
	std::vector<triangle> tris;
	FILE *f=fopen("cat_lopoly.obj","r");
	char line[1000];
	while (NULL!=fgets(line,1000,f)) {
		switch (line[0]) {
		case 'v': {
			float x,y,z;
			sscanf(line,"v%f%f%f",&x,&y,&z);
			vec3 p=0.5*vec3(x,y,z);
			texcoords.push_back(p);
			p.z+=0.02*sin(10.0*p.x + time_uniform); // rippling fur
			points.push_back(p);
			normals.push_back(vec3(0.0,0.0,0.0));
			break;
		}
		case 'f': {
			int a[3];
			sscanf(line,"f%d%d%d",&a[0],&a[1],&a[2]);
			vec3 N=normalize(cross(points[a[0]]-points[a[1]],
				points[a[0]]-points[a[2]]));
			for (int i=0;i<3;i++) normals[a[i]]+=N;
			tris.push_back(triangle(a));
			break;
		}
		default: /* ignore that line */ ;
		}
	}
	fclose(f);
	
	/* normalize the normals */
	for (int n=1;n<(int)normals.size();n++) normals[n]=normalize(normals[n]);
	
	/* Draw many alpha-blended layers of fur */
	glEnable(GL_DEPTH_TEST);
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
	float furdist=0.2;
	for (float fur=0.0;fur<=1.0;fur+=1.0/32.0)  //<- looks much better with 32 layers
	{
		if (fur>0.0) glDepthMask(GL_FALSE);
	
		glBegin(GL_TRIANGLES); 
		for (int t=0;t<(int)tris.size();t++)
		{
			for (int i=0;i<3;i++) {
				int vertex=tris[t].a[i];
				// The normal is used for lighting
				glNormal3fv(normals[vertex]);

				// The texcoords determine the hair roots
				vec3 p=texcoords[vertex];
				glTexCoord4fv(vec4(p.x,p.y,p.z,fur));

				// The geometry shows where stuff lands onscreen
				p=points[vertex];
				p+=(fur*furdist)*normals[vertex];
				glVertex3fv(p);
			}
		}
		glEnd();
	}
	glDepthMask(GL_TRUE);
	
	glUseProgramObjectARB(0);
	
	glutPostRedisplay(); // continual animation
	glutSwapBuffers();
	
/* Estimate framerate */
	static double last_time=0.0;
	static int framecount=0, last_framecount=0;
	framecount++;
	double cur_time=0.001*glutGet(GLUT_ELAPSED_TIME);
	if (cur_time>0.5+last_time) 
	{ // every half a second, estimate fps
		double time=cur_time-last_time;
		double time_per_frame=time/(framecount-last_framecount);
		char str[100];
		sprintf(str,"Solid Texture Renderer: %.1f fps, %.1f ms/frame",
			1.0/time_per_frame,1.0e3*time_per_frame);
		glutSetWindowTitle(str);
		last_framecount=framecount;
		last_time=cur_time;
	}
}

int main(int argc,char *argv[]) 
{
	glutInit(&argc,argv);
	glutInitDisplayMode(GLUT_RGBA + GLUT_DEPTH + GLUT_DOUBLE);
	glutInitWindowSize(800,600);
	glutCreateWindow("Demo Application");
	
	glewInit();
	
	glutDisplayFunc(display);
	oglCameraInit();
	camera=vec3(0.0,-2.0,0.0);
	correct_head_tilt=true; // allow free flight
	
	glutMainLoop();
	return 0;
}
